#!/usr/bin/python3

from typing import Dict, Iterable, Optional, Tuple
from configparser import ConfigParser 
from uuid import uuid1
from locale import atoi
import os, glob, json, logging

from bottle import route, run, response, request, static_file, HTTPError, HTTPResponse
from xml.dom.minidom import parseString
import xml.etree.ElementTree as ET


restServiceHost = '0.0.0.0'
restServicePort = 28080
ttnList = []
ttnListV3 = []
ttnListV4 = []
formbList = []
callCount = {}
testMode = False
ruleList: Dict[str, Iterable[str]] = {}
exciseBottleAttrs = ['price', 'barcode', 'ean', 'volume']
nonExciseBottleAttrs = ['code', 'price', 'bname', 'alc', 'volume', 'ean', 'count']
replyID = ''
logging.basicConfig(format=u'%(levelname)-8s [%(asctime)s] %(message)s', level=logging.DEBUG, filename=u'egaistest.log')


def getRule(funcName: str) -> Optional[str]:
    rule = None
    func = funcName.lower()
    if testMode and func in ruleList:
        if func not in callCount:
            callCount.update({func: 0})
        callCount[func] += 1

        curNumber = callCount[func]
        if curNumber >= len(ruleList[func]) or ruleList[func][curNumber] == None:
            curNumber = 0
        rule = ruleList[func][curNumber]
    return rule


@route('/', method='GET')
def checkServerAvailable() -> Tuple[str]:
    logging.info(u'Запрос доступности сервера')
    rule = getRule('checkServerAvailable')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    response.content_type = 'text/xml; charset=utf8'
    logging.info(u'Сервер доступен')
    return ("<?xml version='1.0' encoding='UTF-8' standalone='no'?>")


@route('/exciseMark/add', method='POST')
def exciseMarkAdd() -> None:
    '''
    Прием данных о продаже алькогольной продукции с акцизными марками
    '''
    logging.info(u'Принимаем данные о продаже товара с акцизной маркой')
    logging.info(request.body.read())
    logging.info(u'Данные приняли')


@route('/xml', method='POST')
def getEgaisCheck() -> HTTPResponse:
    logging.info(u'Получаем данные для ЕГАИС')
    rule = getRule('getEgaisCheck')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Получен пустой чек')
        raise HTTPError(204)
    elif rule == 'incorrectCheque':
        logging.error(u'Некорректные данные чека')
        raise HTTPError(400)
    elif rule == 'fieldRequired':
        logging.error(u'Отсутствует необходимый атрибут')
        raise HTTPError(400)
    elif rule == 'exciseDuplication':
        logging.error(u'Дублирование акцизной позиции')
        raise HTTPError(400)
    elif rule == 'checkFailed':
        logging.error(u'Акцизная марка не прошла проверку ЕГАИС')
        data = None
        with open('xml/egais_answer_check.xml', 'rb') as f:
            data = f.read()
        raise HTTPResponse(data, content_type = 'text/xml; charset=utf8', status=500)
    xml = request.forms.get("xml_file")
    try:
        root = ET.fromstring(xml)

        excise = root.findall('Bottle')
        if len(excise) > 0:
            logging.info(u'Проверка акцизной продукции')
            checkExcise(excise)

        nonExcise = root.findall('nopdf')
        if len(nonExcise) > 0:
            logging.info(u'Проверка неакцизной продукции')
            checkNonExcise(nonExcise)
    except Exception as e:
        logging.error(e)
        messageIncorrectData = 'Некорректные данные чека: cheque.xml'
        logging.warning(messageIncorrectData)
        raise HTTPError(400, messageIncorrectData)
    logging.info(u'Отправляем ответ от ЕГАИС')
    return static_file('egais_answer.xml', root='xml')


# Проверка неакцизной продукции на наличие обязательных полей
# Проверка наличия значений и типов не производится
def checkNonExcise(xmlNodesList: Iterable[ET.Element]) -> None:
    for xmlNode in xmlNodesList:
        for attr in nonExciseBottleAttrs:
            if xmlNode.attrib.get(attr) == None:
                raise HTTPError(400,
                                'В неакцизной продукции %s' % xmlNode | ' отсутствует необходимый атрибут: %s' % attr)

            # Проверка акцизной продукции на наличие обязательных полей


# Проверка наличия значений и типов не производится,
# кроме проверки уникальности поля 'barcode'
def checkExcise(xmlNodesList: Iterable[ET.Element]) -> None:
    barcodes = {}
    for xmlNode in xmlNodesList:
        for attr in exciseBottleAttrs:
            if xmlNode.attrib.get(attr) == None:
                raise HTTPError(400,
                                'В акцизной продукции %s' % xmlNode | ' отсутствует необходимый атрибут: %s' % attr)
            if attr == 'barcode':
                barcodes[xmlNode.attrib.get(attr)] = 1
    if len(xmlNodesList) != len(barcodes):
        raise HTTPError(400, 'Обнаружено дублирование акцизной продукции')


@route('/opt/out', method='GET')
def sendDocList() -> bytes:
    logging.info(u'Получили запрос на список документов')
    rule = getRule('sendDocList')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Возвращаем пустое сообщение')
        raise HTTPError(204)
    elif rule == 'nottn':
        response.content_type = 'text/xml; charset=utf8'
        logging.error(u'Возвращаем пустой список')
        return ET.tostring(ET.Element('A'))

    ttnList.sort()
    ttnListV3.sort()
    ttnListV4.sort()
    formbList.sort()

    root = ET.Element('A')
    for ttn in ttnList:
        child = ET.SubElement(root, 'url')
        child.text = 'http://' + str(restServiceHost) + ':' + str(restServicePort) + '/opt/out/WAYBILL_v2/' + ttn
    for ttn in ttnListV3:
        child = ET.SubElement(root, 'url')
        child.text = 'http://' + str(restServiceHost) + ':' + str(restServicePort) + '/opt/out/WAYBILL_v3/' + ttn
    for ttn in ttnListV4:
        child = ET.SubElement(root, 'url')
        child.text = 'http://' + str(restServiceHost) + ':' + str(restServicePort) + '/opt/out/WAYBILL_v4/' + ttn

    for formb in formbList:
        child = ET.SubElement(root, 'url')
        child.text = 'http://' + str(restServiceHost) + ':' + str(restServicePort) + '/opt/out/FORM2REGINFO/' + formb
    s = ET.tostring(root)
    logging.info(u'Возвращаем список документов %s' % s)
    response.content_type = 'text/xml; charset=utf8'
    return s


@route('/opt/out/WAYBILL_v4/<filename>', method='GET')
def sendTTN(filename: str) -> HTTPResponse:
    logging.info(u'Получили запрос на получение ТТН V4')
    rule = getRule('sendTTN')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Возвращаем пустое сообщение')
        raise HTTPError(204)
    logging.info(u'Возвращаем ТТН V4 %s.xml' % filename)
    return static_file(filename + '.xml', root='ttnV4')

@route('/opt/out/WAYBILL_v3/<filename>', method='GET')
def sendTTN(filename: str) -> HTTPResponse:
    logging.info(u'Получили запрос на получение ТТН V3')
    rule = getRule('sendTTN')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Возвращаем пустое сообщение')
        raise HTTPError(204)
    logging.info(u'Возвращаем ТТН V3 %s.xml' % filename)
    return static_file(filename + '.xml', root='ttnV3')


@route('/opt/out/WAYBILL_v2/<filename>', method='GET')
def sendTTN(filename: str) -> HTTPResponse:
    logging.info(u'Получили запрос на получение ТТН')
    rule = getRule('sendTTN')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Возвращаем пустое сообщение')
        raise HTTPError(204)
    logging.info(u'Возвращаем ТТН %s.xml' % filename)
    return static_file(filename + '.xml', root='ttn')


@route('/opt/out/FORM2REGINFO/<filename>', method='GET')
def sendFormb(filename: str) -> HTTPResponse:
    logging.info(u'Получили запрос на получение уведомления')
    rule = getRule('sendFormb')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Возвращаем пустое сообщение')
        raise HTTPError(204)
    logging.info(u'Возвращаем уведомление %s.xml' % filename)
    return static_file(filename + '.xml', root='form2')


@route('/opt/in/WayBillAct_v2', method='POST')
@route('/opt/in/WayBillAct_v3', method='POST')
@route('/opt/in/WayBillAct_v4', method='POST')
def recieveAct() -> bytes:
    logging.info(u'Принимаем акт')
    rule = getRule('recieveAct')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'fieldRequired':
        logging.error(u'Отсутствует поле WBRegId')
        raise HTTPError(400, "Required field 'WBRegId'")

    xml = request.forms.get("xml_file")
    dom = parseString(xml)
    elem = dom.getElementsByTagName('wa:WBRegId')
    if elem and elem[0].childNodes:
        regId = elem[0].childNodes[0].data
    else:
        logging.error(u'Отсутствует или не заполнено поле WBRegId')
        raise HTTPError(400, "Required field 'WBRegId'")

    # xmlschema = etree.XMLSchema(etree.parse("xsd/act.xsd"))
    # xx = etree.fromstring(xml)
    # xmlschema.assertValid(xx)
    actName = 'act%s.xml' % regId
    if not os.path.exists('act'):
        os.mkdir('act')
    with open(os.path.join('act', actName), "w") as f:
        logging.info(u'Акт %s сохранен ' % actName)
        f.write(xml)

    root = ET.Element('A')
    if replyID:
        ET.SubElement(root, 'url').text = replyID
    else:
        ET.SubElement(root, 'url').text = str(uuid1())
    ET.SubElement(root,
                  'sign').text = '895B92CAD115B57B02C7D12ADC488066D99B60549D57A737B0CAC18E5E3E1C72E6B8D414C763CB58A5E67DE7C8C2ECE908451C4AE6838479A42ABBA8179D0CE8'
    ET.SubElement(root, 'ver').text = '2'

    ans = ET.tostring(root)
    logging.info(u'Возвращаем ответ на акт %s' % ans)
    response.content_type = 'text/xml; charset=utf8'

    return ans


@route('/opt/out/WAYBILL_v2/<filename>', method='DELETE')
def deleteTTN(filename: str) -> None:
    logging.info(u'Запрос на удаление ТТН %s' % filename)
    rule = getRule('deleteTTN')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    ttnList.remove(filename)
    logging.info(u'ТТН удалена')


@route('/opt/out/FORM2REGINFO/<filename>', method='DELETE')
def deleteFormb(filename: str) -> None:
    logging.info(u'Запрос на удаление уведомления %s' % filename)
    rule = getRule('deleteFormb')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    formbList.remove(filename)
    logging.info(u'Уведомление удалено')


@route('/exciseMark/status', method='PUT')
def sendExciseMarkStatus() -> None:
    logging.info(u'Отправка статуса акцизных марок')
    body = json.loads(request.body.read())
    logging.info(u'Список акцизных марок: %s' % body)
    rule = getRule('sendExciseMark')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)


@route('/exciseMark/status', method='POST')
def checkExciseMark() -> str:
    logging.info(u'Проверка акцизных марок')
    rule = getRule('checkExciseMark')
    if rule == 'errorConnection':
        logging.error(u'Ошибка соединения')
        raise HTTPError(404)
    elif rule == 'emptyMessage':
        logging.error(u'Список марок пуст')
        response.status = 400
        response.content_type = 'application/json'
        return json.dumps({"errorCode": 400, "url": "/exciseMark/status", "message": u'Список марок пуст',
                           "errorClassName": "Exception"})

    body = json.loads(request.body.read())
    answer = []
    for mark in body:
        markStatus = {}
        markStatus["exciseMarkId"] = mark
        # цифровой статус
        if rule and rule[0].isdigit():
            markStatus["status"] = atoi(rule)
        # марка не зарегистрирована в БД
        elif rule == 'notRegisteredInDatabase':
            markStatus["status"] = 0
        # товар с такой маркой продан
        elif rule == 'goodsSold':
            markStatus["status"] = 1
        # товар с такой маркой возвращен
        elif rule == 'goodsReturned':
            markStatus["status"] = 2
        # марка зарезервирована в чеке продажи
        elif rule == 'reservedInCheckSale':
            markStatus["status"] = 3
        # марка снята с резерва в чеке продажи
        elif rule == 'removedFromReserveInCheckSale':
            markStatus["status"] = 4
        # марка зарезервирована в чеке возврата (чеке возврата по чеку продажи)
        elif rule == 'reservedInCheckReturn':
            markStatus["status"] = 5
        # марка снята с резерва в чеке возврата (чеке возврата по чеку продажи)
        elif rule == 'removedFromReserveInCheckReturn':
            markStatus["status"] = 6
        # марка отсутствует в списке допустимых к продаже
        elif rule == 'goodsInBlackList':
            markStatus["status"] = 7
        # алкокод не найден
        elif rule == 'alcocodeNotFound':
            markStatus["status"] = 8
        # нет на остатке
        elif rule == 'outOfBalance':
            markStatus["status"] = 9
        # не передан штрихкод для проверки в белом списке
        elif rule == 'noBarcodeForCheck':
            markStatus["status"] = 10
        # не найдено соответствие акцизной марки и штрихкода в белом списке
        elif rule == 'noMatchingBarcodeAndExise':
            markStatus["status"] = 11
        # не найден магазин для проверки остатков
        elif rule == 'shopForRemainsCheckNotFound':
            markStatus["status"] = 12
        # не заполнен штрихкод для проверки остатков
        elif rule == 'noBarcodeForRemainsCheck':
            markStatus["status"] = 13
        # не заполнен код товара для проверки остатков
        elif rule == 'noTmcForRemainsCheck':
            markStatus["status"] = 14
        # некорректно заполнен справочник остатков для товара
        elif rule == 'incorrectRemainsCheckDict':
            markStatus["status"] = 15
        # неизвестный кассе статус(необходимо для проверки ошибки)
        elif rule == 'unknownStatus':
            markStatus["status"] = 999
        else:
            markStatus["status"] = 0
        message = getRule('returnMessage')
        if message:
            markStatus["message"] = message
        answer.append(markStatus)
    response.content_type = 'application/json'

    return json.dumps(answer)


def init() -> None:
    logging.info(u'Чтение настроек')
    global restServiceHost
    global restServicePort
    global testMode
    global replyID
    conf = ConfigParser()
    conf.read('conf/config.ini')
    restServiceHost = conf.get('main', 'host')
    restServicePort = conf.getint('main', 'port')
    testMode = conf.getboolean('main', 'testMode')
    replyID = conf.get('main', 'replyID')
    if testMode:
        for key, val in conf.items('test'):
            rule = key.split('.')
            beg = 0
            end = 0
            func = rule[0]
            if not func in ruleList:
                ruleList[func] = []
            if len(rule) > 1:
                interval = rule[1].split('-')
                beg = int(interval[0])
                end = int(interval[0])
                if len(interval) > 1:
                    end = int(interval[1])
            i = beg
            while i <= end:
                while i >= len(ruleList[func]):
                    ruleList[func].append(None)
                ruleList[func][i] = val
                i += 1
        logging.info(u'Заданы правила для сервиса: %s' % ruleList)


def start() -> None:
    logging.info(u'Запускаем сервис')
    try:
        init()
        listing = glob.glob('ttnV4/*.xml')
        for filename in listing:
            ttnListV4.append(os.path.splitext(os.path.basename(filename))[0])
        listing = glob.glob('ttn/*.xml')
        for filename in listing:
            ttnList.append(os.path.splitext(os.path.basename(filename))[0])
        listing = glob.glob('ttnV3/*.xml')
        for filename in listing:
            ttnListV3.append(os.path.splitext(os.path.basename(filename))[0])
        listing = glob.glob('form2/*.xml')
        for filename in listing:
            formbList.append(os.path.splitext(os.path.basename(filename))[0])
        logging.info(u'Стартуем на %s:%s' % (restServiceHost, restServicePort))
        run(host=restServiceHost, port=restServicePort)
    except Exception as e:
        print(e)


if __name__ == '__main__':
    start()
